/****************************************************************************

    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#include "Newtonator2.h"
#include "config.h"
#include <math.h>
#include <algorithm>
#include <string.h>
#include <stdlib.h>

using namespace stk;

Newtonator2::Newtonator2(unsigned int channels)
	: MAX_VAL(1.0), MIN_VAL(-1.0)
	, _cpGModFreqMult(2, 0, 1)
	, _cpWobbleFreq(0, 0, 1)
	, _cpMainAmp(0, 2, 0)
{
	_set = NULL;
    _pPrevPos = NULL;
    _pPrevVeloc = NULL;
    _pHoldVeloc = NULL;
    _gravDir = NULL;
    _gravModSource = NULL;
    _velocDelay = NULL;
    _isClipping = NULL;
    _wobbleWav = NULL;
    _a = NULL;
    _gravity = 0.0;
    _noteCurrMaxRM = 0.0;
    _currVelocSaH = 0;
    _ampEnv = NULL;
    _numChannels = 0;
    _noteAmp = 0.0;
    _noteFrequency = 0.0;
    _sampCount = 0;
    _timeForCycle = 0.0;
    _needsKeyOff = false;
    initChannels(channels);
}

Newtonator2::~Newtonator2()
{
    clearChannelMemory();
}

void Newtonator2::setSettings(INewtSettings* set)
{
    _set = set;

    if (_gravModSource)
    {
        for (unsigned int i = 0; i < _numChannels; i++)
        {
            _gravModSource[i].setSettings(_set);
        }
    }
}

CeilingBehavEnum Newtonator2::GetCBFromFloat(stk::StkFloat val)
{
    switch ((int)val)
    {
        case 0:
            return ExtClip;
        case 1:
            return IntClip;
        case 2:
            return Bounce;
        case 3:
            return BounceMirror;
        default:
            return Unknown;
    }
}

void Newtonator2::clearChannelMemory()
{
    if (_pPrevPos != NULL) delete [] _pPrevPos;
    if (_pPrevVeloc != NULL) delete [] _pPrevVeloc;
    if (_pHoldVeloc != NULL) delete [] _pHoldVeloc;
    if (_gravDir != NULL) delete [] _gravDir;
    if (_gravModSource != NULL) delete [] _gravModSource;
    if (_ampEnv != NULL) delete [] _ampEnv;
    if (_isClipping != NULL) delete [] _isClipping;
    if (_wobbleWav != NULL) delete [] _wobbleWav;
    if (_a != NULL) delete [] _a;
    if (_velocDelay != NULL) delete [] _velocDelay;
    _pPrevPos = NULL;
    _pPrevVeloc = NULL;
    _pHoldVeloc = NULL;
    _gravDir = NULL;
    _gravModSource = NULL;
    _ampEnv = NULL;
	_set = NULL;
	_velocDelay = NULL;
	_isClipping = NULL;
	_wobbleWav = NULL;
	_a = NULL;
    _numChannels = 0;
}

void Newtonator2::initChannels(unsigned int numChannels)
{
    //If the number of channels has changed, reset all channel
    //memory variables.
    if (numChannels != _numChannels && numChannels > 0)
    {
        clearChannelMemory();
        _pPrevPos = new StkFloat[numChannels];
        _pPrevVeloc = new StkFloat[numChannels];
        _pHoldVeloc = new StkFloat[numChannels];
        _gravDir = new int[numChannels];
        _gravModSource = new VariSource[numChannels];
        _ampEnv = new ADSR[numChannels];
        _velocDelay = new Delay[numChannels];
        _isClipping = new bool[numChannels];
        _wobbleWav = new SineWave[numChannels];
        _a = new StkFloat[numChannels];
        _numChannels = numChannels;

        for (unsigned int i = 0; i < _numChannels; i++)
        {
            _pPrevPos[i] = MAX_VAL;
            _pPrevVeloc[i] = 0.0;
            _pHoldVeloc[i] = 0.0;
            _gravDir[i] = -1;
            _gravModSource[i].setUsingEnvelope(false);
            _gravModSource[i].setSettings(_set);
            _gravModSource[i].setUse(VariSource::GMOD);
            _velocDelay[i].setMaximumDelay(100);
            _isClipping[i] = false;
        }
    }
}

void Newtonator2::keyOn(stk::StkFloat velocity)
{
    TRACE1("keyOn", this);
    _needsKeyOff = true;
    for (unsigned int chan = 0; chan < numChannels(); chan++)
    {
        _ampEnv[chan].keyOn();
        _noteAmp = velocity;
        if (_gravModSource[chan].isUsingEnvelope())
        {
            _gravModSource[chan].setAmplitude(velocity);
            _gravModSource[chan].getEnv().keyOn();
        }
    }
}

void Newtonator2::keyOff(stk::StkFloat velocity)
{
    TRACE1("keyOff", this);
    for (unsigned int chan = 0; chan < numChannels(); chan++)
    {
        _ampEnv[chan].keyOff();

        if (_gravModSource[chan].isUsingEnvelope())
            _gravModSource[chan].getEnv().keyOff();
    }
    _needsKeyOff = false;
}

bool Newtonator2::isPlaying() const
{
    for (unsigned int chan = 0; chan < numChannels(); chan++)
    {
        if (_ampEnv[chan].getState() != ADSR::IDLE) return true;
    }

    return false;
}

void Newtonator2::setFrequency(stk::StkFloat frequency)
{
    //Set the gravity accordingly
    resetForFreqChange();
    _timeForCycle = Stk::sampleRate() / frequency;
    StkFloat timeForFreq = _timeForCycle / 4;
    StkFloat teeSquared = pow(timeForFreq, 2);

    _gravity = (2 * MAX_VAL) / teeSquared;

    TRACE1("setFrequency, frequency", frequency);

    //So that we know what the frequency is supposed to be
    //for this newt...
    _actualFrequency = _noteFrequency = frequency;
}

void Newtonator2::bendFrequency(stk::StkFloat factor)
{
    _actualFrequency = _noteFrequency * factor;
    _timeForCycle = Stk::sampleRate() / _actualFrequency;

    StkFloat timeForFreq = _timeForCycle / 4;
    StkFloat teeSquared = pow(timeForFreq, 2);

    _gravity = (2 * MAX_VAL) / teeSquared;
}

void Newtonator2::resetForFreqChange()
{
    _sampCount = 0;
    _noteCurrMaxRM = 0.0;
    _currVelocSaH = 0;
    for (unsigned int i = 0; i < _numChannels; i++)
    {
        _pPrevPos[i] = MAX_VAL * (1 - ((random() / 20) * getChanSep(i) / RAND_MAX));
        _pPrevVeloc[i] = 0.0;
        _pHoldVeloc[i] = 0.0;
        _gravDir[i] = -1;
        _isClipping[i] = false;
        _velocDelay[i].clear();
        _gravModSource[i].reset();
    }
}

void Newtonator2::syncSynthSettings()
{
	for (unsigned int chan = 0; chan < numChannels(); chan++)
    {
		//Amplitude Envelope
		_ampEnv[chan].setAllTimes(_set->getAmpAttack(chan), _set->getAmpDecay(chan), _set->getAmpSustain(chan), _set->getAmpRelease(chan));

		//Gravity Modulator
		bool usingGModEnv = _set->getGModEnv(chan);
		_gravModSource[chan].setUsingEnvelope(usingGModEnv);
        if (usingGModEnv)
		{
            _gravModSource[chan].getEnv().setAllTimes(_set->getGModAttack(chan), _set->getGModDecay(chan), _set->getGModSustain(chan), _set->getGModRelease(chan));
        }
        if (_set->getGModFollowFreq(chan))
        {
            _gravModSource[chan].setFrequency(_actualFrequency * (1 + _set->getGModFreqDrift(chan)) * _cpGModFreqMult.Compute(2, _set->getGModFreqDiv(chan)));
        }
        else
        {
            _gravModSource[chan].setFrequency(_set->getGModFrequency(chan));
        }
		_gravModSource[chan].setType(_set->getGModType(chan));
    }
}

StkFloat Newtonator2::limitVal(StkFloat val, StkFloat limit)
{
    if (val > 0)
        return std::min(limit, val);
    else
        return std::max(-limit, val);
}

StkFloat Newtonator2::procVeloc(StkFloat veloc, unsigned int chan)
{
    //Process the velocity and return the velocity value to store
    //for the next newtTick.
    veloc = doVelocSaH(veloc, chan);
    veloc = doVelocDelay(veloc, chan);

    return veloc;
}

stk::StkFloat Newtonator2::doVelocDelay(stk::StkFloat veloc, unsigned int chan)
{
    int delay = (int)getVelocDelay(chan);
    StkFloat ret = veloc;
    StkFloat ratio = (StkFloat)delay / 100.0;

    if (delay >= 1)
    {
        _velocDelay[chan].setDelay((unsigned int)delay);
        ret = _velocDelay[chan].tick(veloc) + (ratio * veloc);
    }

    return ret;
}

stk::StkFloat Newtonator2::doVelocSaH(stk::StkFloat veloc, unsigned int chan)
{
    //Samp & Hold will scale with the note frequency; see getVelocSaH impl.
    int sah = (int)getVelocSaH(chan);
    StkFloat ret = veloc;

    if (sah >= 2)
    {
        ret = _pHoldVeloc[chan];

        if (++_currVelocSaH >= sah * numChannels())
        {
            _currVelocSaH = 0;
            _pHoldVeloc[chan] = veloc;
        }
        else if (getStereo() && _currVelocSaH >= (sah * numChannels()) - 1)
        {
            _pHoldVeloc[chan] = veloc;
        }
    }

    return ret;
}

StkFloat Newtonator2::separate(StkFloat val, unsigned int chan) const
{
    return (val * (1 + (const_cast<Newtonator2*>(this)->getChanSep(chan) * (chan == 0 ? -1 : 1))));
}

void Newtonator2::resetWobble(unsigned int chan)
{
    StkFloat wobFreq = 1000.0 + (_cpWobbleFreq.Compute(getWobbleTimbre(chan), 2.0) * 1000000.0);
    _wobbleWav[chan].reset();
    _wobbleWav[chan].setFrequency(std::abs(_pPrevVeloc[chan] * wobFreq));
    _a[chan] = getWobbleA(chan);
}

StkFloat Newtonator2::wobbleClip(StkFloat currSample, StkFloat modifier, unsigned int chan)
{
    StkFloat scale = (1.0 - getWobbleScale(chan));

    if (!_isClipping[chan])
        return currSample * scale * modifier;

    StkFloat retSample = 0;
    StkFloat sine = _wobbleWav[chan].tick();
    StkFloat clip = (currSample > 0 ? getCeilingTop(chan) : getCeilingBottom(chan));

    retSample = clip + (_pPrevVeloc[chan] * _a[chan] * sine * getWobbleScale(chan) * 8);
    _a[chan] -= (getWobbleB(chan) * _a[chan]);

    return retSample * scale * modifier;
}

StkFloat Newtonator2::ringMod(StkFloat carrier, StkFloat mod, unsigned int chan)
{
    StkFloat amount = getVelocRM(chan);

    if (amount > 0.0)
    {
        return carrier * (1 - ((mod / _noteCurrMaxRM) * amount));
    }

    return carrier;
}

void Newtonator2::velocResetForSampAndHold(unsigned int channel, StkFloat velocReset)
{
    if (getVelocSaH(channel) > _timeForCycle && velocReset < 1 && velocReset > -1)
    {
        //This would introduce a condition whereby the velocity reset could
        //diminish the practical velocity to zero for the length of the frequency cycle.
        //Try to mitigate this as much as possible.
        StkFloat sign = (velocReset >= 0 ? 1 : -1);
        _pHoldVeloc[channel] *= (sqrt(std::abs(velocReset)) * sign);
    }
    else
    {
        _pHoldVeloc[channel] *= velocReset;
    }
}

StkFloat Newtonator2::newtTick(unsigned int channel, unsigned int frameIdx)
{
    StkFloat retSample = 0;

    if (_numChannels == 0) return 0.0;

	//Synchronize external synth settings with internal member variables
	syncSynthSettings();

    //Initiate channel "memory" of floor and positions
    StkFloat dz = getDeadZone(channel);
    StkFloat& prevPos = _pPrevPos[channel];
    StkFloat prevVeloc = _pPrevVeloc[channel];
    int& gravDir = _gravDir[channel];
    StkFloat floorVal = 0.0;
    StkFloat deadZoneA = floorVal + dz;
    StkFloat deadZoneB = floorVal - dz;
    StkFloat gravMod = getGravModDepth(channel) * _gravModSource[channel].tick(frameIdx, channel);
    StkFloat gravity = _gravity * (getGModDCOffset(channel) + gravMod);
    StkFloat tempPos;
    StkFloat tempVeloc = prevVeloc;
    StkFloat gravScale = getGravityScale(channel);
    StkFloat currGrav = gravDir * gravity;
    StkFloat currVeloc = 0.0;

    //Main computation
    if (prevPos > deadZoneB && prevPos < deadZoneA)
    {
        //Gravity has no effect in deadzone
        tempPos = prevPos + procVeloc(prevVeloc, channel);
    }
    else
    {
        tempPos = prevPos + procVeloc(prevVeloc, channel) + ((currGrav + (prevVeloc * -gravScale)) * .5);
        tempVeloc += currGrav;
    }

    //check to see if signal has passed through floor
    if ((gravDir < 0 && tempPos < deadZoneB) || (gravDir > 0 && tempPos > deadZoneA))
    {
        gravDir *= -1;
    }

    //check to do velocity reset
    StkFloat velocReset = getVelocReset(channel);
    if (_sampCount >= (_timeForCycle * numChannels()) + 1)
    {
        //Make sure we're not degrading the velocity in the dead zone, because if we do, the signal could just die
        //if it never makes it past the deadzone boundaries
        if (prevPos < deadZoneB || prevPos > deadZoneA || velocReset <= -1)
        {
            tempVeloc *= velocReset;

            //do it for sample & hold too
            velocResetForSampAndHold(channel, getVelocReset(channel));
        }
        _sampCount = _sampCount - floor(_sampCount);
    }
    else if (_sampCount >= _timeForCycle * numChannels())
    {
        if (prevPos < deadZoneB || prevPos > deadZoneA || velocReset <= -1)
        {
            tempVeloc *= velocReset;

            //do it for sample & hold too
            velocResetForSampAndHold(channel, getVelocReset(channel));
        }
        if (!getStereo()) _sampCount = _sampCount - floor(_sampCount);
    }

    //See if the signal is clipping
    if (tempPos < getCeilingBottom(channel) || tempPos > getCeilingTop(channel))
    {
        if (!_isClipping[channel])
            resetWobble(channel);
        _isClipping[channel] = true;
    }
    else
    {
        _isClipping[channel] = false;
    }

    //ceiling behavior
    switch (getCeilingBehavior(channel))
    {
        case IntClip:
            if (tempPos >= 0)
            {
                retSample = std::min(getCeilingTop(channel), tempPos);
            }
            else if (tempPos < 0)
            {
                retSample = std::max(getCeilingBottom(channel), tempPos);
            }
            break;
        case Bounce:
            // Only reverse velocity if we're clipping AND the velocity hasn't been reversed yet
            if (_isClipping[channel] && ((tempPos > deadZoneA && tempVeloc > 0) || (tempPos < deadZoneB && tempVeloc < 0)))
            {
                tempVeloc *= getCeilingBounce(channel);
                _pHoldVeloc[channel] *= getCeilingBounce(channel);
            }
            retSample = tempPos;
            break;
        default:
            retSample = tempPos;
            break;
    }

    //remember channel values
    _pPrevVeloc[channel] = limitVal(tempVeloc, MAX_VAL);
    prevPos = retSample;

    //remember what the maximum velocity encountered is for this note
    currVeloc = std::abs(tempVeloc);
    if (currVeloc > _noteCurrMaxRM) _noteCurrMaxRM = currVeloc;

    //do ring mod with velocity signal
    retSample = ringMod(retSample, tempVeloc, channel);

    //apply note envelope, if applicable
    StkFloat env = _ampEnv[channel].tick();
    retSample = wobbleClip(retSample, _noteAmp * env, channel);

    //ext. limiter
    retSample = limitVal(retSample, getCeilingTop(channel)) * getAmp(channel);

    _sampCount++;

    return retSample;
}
